try:
    from isomut2py import io

    import theano.tensor as __tt
    import numpy as __np
    import pymc3 as __pm
    import pandas as __pd
    from scipy import stats as __stats
    import gc as __gc
    import logging as __logging
except ImportError:
    print('ImportError in isomut2py.ploidyEstBayesian, Bayesian estimate functions will not work.')


def estimate_hapcov_infmix(cov_min=None, hc_percentile=None,
                           chromosomes=None, output_dir=None, cov_max=None, level=0, cov_sample=None,
                           user_defined_hapcov=None):
    """

    Estimates the haploid coverage of the sample.
    If the user_defined_hapcov attribute is set manually, it sets the value of estimated_hapcov to that.
    Otherwise, a many-component (20) Gaussian mixture model is fitted to the coverage histogram of the sample 10 times. Each time,
    the haploid coverage is estimated from the center of the component with the maximal weight in the model. The final estimate of the haploid
    coverage is calculated as the qth percentile of the 10 measurements, with q = hc_percentile.

    :param user_defined_hapcov: if not None, its value is returned (default: None) (float)
    :param cov_sample: a sample of the coverage distribution of the investigated sample, if None, it is loaded from the temporary files of the output_dir (default: None) (array-like)
    :param cov_max: the maximum value of the coverage for a position to be considered in the estimation (default: None) (int)
    :param output_dir: the path to the output directory of the PloidyEstimator object, where temporary files are located (default: None) (str)
    :param chromosomes: list of chromosomes for the sample (default: None) (array-like)
    :param hc_percentile: the percentile value to use for calculating the estimated haploid coverage from 10 subsequent estimations (default: None) (int)
    :param cov_min: the maximum value of the coverage for a position to be considered in the estimation (default: None) (int)
    :param level: the level of indentation used in verbose output (default: 0) (int)

    :returns: (estimated haploid coverage (float), sample from coverage distribution (array-like))

    """

    # if a sample of the coverage distribution is not provided, generate one
    if cov_sample is None:
        cov_sample = io.get_coverage_distribution(chromosomes=chromosomes,
                                                  output_dir=output_dir,
                                                  cov_max=cov_max,
                                                  cov_min=cov_min)

    # if the user has set the haploid coverage manually, use that
    if user_defined_hapcov is not None:
        return user_defined_hapcov, cov_sample


    # fit a finite mixture model in 10 chains with pymc3
    K = 20  # number of mixture components
    number_of_chains = 10  # number of chains to use
    iterations = 15000  # number of iterations
    burn_beginning = 10000  # number of iterations to ignore from the beginning (where it might not have converged yet)

    def stick_breaking(beta):
        portion_remaining = __tt.concatenate([[1], __tt.extra_ops.cumprod(1 - beta)[:-1]])
        return beta * portion_remaining

    def fit_infinite_mixture_model(coverage_dist, K, number_of_chains_to_use, number_of_iterations, burn_period):
        covdist_standard = (coverage_dist - coverage_dist.mean()) / coverage_dist.std()
        N = len(covdist_standard)

        __gc.collect()

        with __pm.Model() as model:
            alpha = __pm.Gamma('alpha', 1., 1.)
            beta = __pm.Beta('beta', 1., alpha, shape=K)
            w = __pm.Deterministic('w', stick_breaking(beta))

            tau = __pm.Gamma('tau', 1., 1., shape=K)
            lambda_ = __pm.Uniform('lambda', 0, 5, shape=K)
            mu = __pm.ExGaussian('mu', mu=-4, sigma=__np.sqrt(1 / (lambda_ * tau)), nu=5, shape=K)
            obs = __pm.NormalMixture('obs', w, mu, tau=lambda_ * tau,
                                     observed=covdist_standard)

        # logger = __logging.getLogger("pymc3")
        # logger.propagate = False

        __logging.getLogger("pymc3").setLevel(__logging.WARNING)

        with model:
            step1 = __pm.Metropolis(vars=[alpha, beta, tau, lambda_, mu], verbose=0)

            tr = __pm.sample(draw=number_of_iterations-burn_period, tune=burn_period,
                             step=[step1], njobs=number_of_chains_to_use,
                             progressbar=False, verbose=0, compute_convergence_checks=False)

        # trace = tr[burn_period:]
        # return trace
        return tr


    trace = fit_infinite_mixture_model(coverage_dist=cov_sample,
                                       K=K,
                                       number_of_chains_to_use=number_of_chains,
                                       number_of_iterations=iterations,
                                       burn_period=burn_beginning)

    chains_to_use = [c for c in range(number_of_chains)]
    hc_all_ests = []
    for c in chains_to_use:
        comp_weights = trace.get_values('w', chains=[c]).mean(axis=0)
        comp_weights_sortidx = __np.argsort(comp_weights)
        standard_means = trace.get_values('mu', chains=[c]).mean(axis=0)[comp_weights_sortidx]
        true_means = standard_means * cov_sample.std() + cov_sample.mean()
        hc_maxweight = true_means[0] / (__np.round(true_means[0] / true_means[true_means > cov_min].min()))
        hc_all_ests.append(hc_maxweight)

    hc_all_ests = __np.array(hc_all_ests)
    hc_maxweight = __np.percentile(a=hc_all_ests[__np.isfinite(hc_all_ests)], q=hc_percentile)

    del trace

    return hc_maxweight, cov_sample


def fit_gaussians(estimated_hapcov,
                  chromosomes=None, output_dir=None, cov_max=None, cov_min=None, level=0, cov_sample=None):
    """

    Fits a 7-component Gaussian mixture model to the coverage distribution of the sample, using the appropriate attributes of the PloidyEstimation
    object. The center of the first Gaussian is initialized from a narrow region around the value of the estimated_hapcov attribute. The centers of
    the other Gaussians are initialized in a region around the value of estimated_hapcov multiplied by consecutive whole numbers.

    The parameters of the fitted model (center, sigma and weight) for all seven Gaussians are both saved to the GaussDistParams.pkl file (in
    output_dir, for later reuse) and set as the value of the distribution_dict attribute.

    :param cov_sample: a sample of the coverage distribution of the investigated sample, if None, it is loaded from the temporary files of the output_dir (default: None) (array-like)
    :param cov_min: the maximum value of the coverage for a position to be considered in the estimation (default: None) (int)
    :param output_dir: the path to the output directory of the PloidyEstimator object, where temporary files are located. If not None, distribution parameters are saved there as GaussDistParams.pkl. (default: None) (str)
    :param chromosomes: list of chromosomes for the sample (default: None) (array-like)
    :param estimated_hapcov: the estimated value for the haploid coverage, used as prior (float)
    :param level: the level of indentation used in verbose output (default: 0) (int)

    :returns: dictionary containing the fitted parameters of the 7 Gaussians

    """

    def get_samples(coverage_distribution, estimated_haploid_cov, number_of_iterations, burn_period):
        K = 7
        halfwidth_of_uniform = 0.2

        __gc.collect()

        model = __pm.Model()
        with model:
            p = __pm.Dirichlet('p', a=__np.array([1., 1., 1., 1., 1., 1., 1.]), shape=K)
            c1 = __pm.Uniform('c1', (1 - halfwidth_of_uniform) * estimated_haploid_cov,
                              (1 + halfwidth_of_uniform) * estimated_haploid_cov)
            means = __tt.stack([c1, c1 * 2, c1 * 3, c1 * 4, c1 * 5, c1 * 6, c1 * 7])
            order_means_potential = __pm.Potential('order_means_potential',
                                                   __tt.switch(means[1] - means[0] < 0, -__np.inf, 0)
                                                   + __tt.switch(means[2] - means[1] < 0, -__np.inf, 0))
            sds = __pm.Uniform('sds', lower=0, upper=estimated_haploid_cov / 2, shape=K)
            category = __pm.Categorical('category',
                                        p=p,
                                        shape=len(coverage_distribution))
            points = __pm.Normal('obs',
                                 mu=means[category],
                                 sd=sds[category],
                                 observed=coverage_distribution)
        with model:
            step1 = __pm.Metropolis(vars=[p, sds, means])
            step2 = __pm.ElemwiseCategorical(vars=[category], values=[0, 1, 2, 3, 4, 5, 6])

            __logging.getLogger("pymc3").setLevel(__logging.WARNING)

            tr = __pm.sample(draw=number_of_iterations-burn_period, tune=burn_period,
                             step=[step1, step2], progressbar=False, verbose=0, compute_convergence_checks=False)
        # trace = tr[burn_period:]
        # return trace
        return tr

    if cov_sample is None:
        cov_sample = io.get_coverage_distribution(chromosomes=chromosomes,
                                                  output_dir=output_dir,
                                                  cov_max=cov_max,
                                                  cov_min=cov_min)

    iterations2 = 15000
    burn_beginning2 = 10000

    # logger = __logging.getLogger("pymc3")
    # logger.propagate = False

    trace2 = get_samples(coverage_distribution=cov_sample,
                         estimated_haploid_cov=estimated_hapcov,
                         number_of_iterations=iterations2,
                         burn_period=burn_beginning2)

    std_trace = trace2.get_values('sds', chains=[0])
    p_trace = trace2.get_values('p', chains=[0])
    sigma = std_trace.mean(axis=0)
    p = p_trace.mean(axis=0)
    mu = __np.array([trace2.get_values('c1', chains=[0]).mean() * (i + 1) for i in range(7)])

    prior_dict = {'mu': mu, 'sigma': sigma, 'p': p}

    del trace2

    if output_dir:
        io.save_obj(prior_dict, output_dir + '/GaussDistParams')

    return prior_dict


def PE_on_chrom(chrom, output_dir, windowsize_PE, shiftsize_PE, distribution_dict):
    """

    Runs the whole ploidy estimation pipeline on a given chromosome, using the appropriate attributes of the PloidyEstimation object by running
    PE_on_range() multiple times.
    Prints the results to the file: [output_dir]/PE_fullchrom_[chrom].txt.

    :param distribution_dict: a dictionary containing the fitted parameters of the Gaussian mixture model to the coverage distribution. (dict with keys 'mu', 'sigma' and 'p')
    :param shiftsize_PE: shiftsize for moving average over regions (int)
    :param windowsize_PE: windowsize for moving average over regions (int)
    :param output_dir: the path to the directory where temporary files are located (str)
    :param chrom: the name of the chromosome to analyse (str)

    """

    if distribution_dict == None:
        raise ValueError('Error: fitted distribution parameters are not defined, cannot proceed.')

    df = __pd.read_csv(output_dir + '/' + 'PEtmp_fullchrom_' + chrom + '.txt', sep='\t',
                       names=['chrom', 'pos', 'cov', 'mut_freq', 'GC']).sort_values(by='pos')

    df['chrom'] = df['chrom'].apply(str)
    pos_all = __np.array(list(df['pos']))
    total_ploidy_a = __np.array([0] * len(pos_all))
    total_loh_a = __np.array([0] * len(pos_all))
    est_num_a = __np.array([0] * len(pos_all))
    posstart = df['pos'].min()
    posmax = df['pos'].max()
    while (posstart < posmax):
        p, loh = PE_on_range(dataframe=df, rmin=posstart, rmax=posstart + windowsize_PE,
                             all_mu=distribution_dict['mu'],
                             all_sigma=distribution_dict['sigma'],
                             prior=distribution_dict['p'])
        if (p > 0):
            total_ploidy_a += p * (pos_all >= posstart) * (pos_all <= posstart + windowsize_PE)
            total_loh_a += loh * (pos_all >= posstart) * (pos_all <= posstart + windowsize_PE)
            est_num_a += 1 * (pos_all >= posstart) * (pos_all <= posstart + windowsize_PE)
        posstart += shiftsize_PE
    df['total_ploidy'] = total_ploidy_a
    df['total_loh'] = total_loh_a
    df['est_num'] = est_num_a

    df['ploidy'] = (df['total_ploidy'] / df['est_num']).round()
    df['LOH'] = (df['total_loh'] / df['est_num']).round()

    df = df[(~df['ploidy'].isnull()) & (~df['LOH'].isnull())]
    df['ploidy'] = df['ploidy'].astype(int)
    df['LOH'] = df['LOH'].astype(int)

    df[['chrom', 'pos', 'cov', 'mut_freq', 'ploidy', 'LOH']].to_csv(
        output_dir + '/PE_fullchrom_' + chrom + '.txt', sep='\t', index=False)
    # return df[['chrom', 'pos', 'cov', 'mut_freq', 'ploidy', 'LOH', 'total_ploidy', 'est_num']]


def PE_on_range(dataframe, rmin, rmax, all_mu, all_sigma, prior, cov_min=0, cov_max=100000):
    """

    Run the ploidy estimation on a given range of a chromosome.

    :param dataframe: The dataframe read from the temporary files for a given chromosome, containing information about the genomic position, the measured coverage and the frequency of the non-reference bases aligned to the position. (pandas.DataFrame)
    :param rmin: The lower bound of the genomic range considered for the analysis. (int)
    :param rmax: The upper bound of the genomic range considered for the analysis. (int)
    :param all_mu: The centers of the seven Gaussians fitted to the coverage distribution. (list of float)
    :param all_sigma: The sigmas of the seven Gaussians fitted to the coverage distribution. (list of float)
    :param prior: The weights of the seven Gaussians fitted to the coverage distribution. (list of float in the range(0,1))

    :returns: (most_probable_ploidy, most_probable_LOH)

        - most_probable_ploidy: the estimated ploidy for the genomic region (int, in range(1,8))
        - most_probable_LOH: the estimated LOH status for the genomic region (1: LOH, 0: no LOH)

    """

    temp = dataframe[(dataframe['pos'] >= rmin) & (dataframe['pos'] <= rmax)
                     & (dataframe['cov'] >= cov_min) & (dataframe['cov'] <= cov_max)]

    if (temp.shape[0] == 0 or temp['pos'].max() == temp['pos'].min()):
        return 0, 0

    # get cov mean
    cov_mean = temp['cov'].mean()

    posteriors = []
    for ploidy in range(len(all_mu)):
        likelihood = __stats.norm.pdf(cov_mean, loc=all_mu[ploidy], scale=all_sigma[ploidy])
        posteriors.append(likelihood * prior[ploidy])
    posteriors = __np.array(posteriors)
    most_probable_ploidy = __np.argmax(posteriors) + 1

    if most_probable_ploidy == 1 or most_probable_ploidy == 3:
        return most_probable_ploidy, 0

    mf_list = __np.array(temp[temp['mut_freq'] >= 0.5]['mut_freq'])
    if len(mf_list) == 0:
        return most_probable_ploidy, 0

    else:
        check_freqs = {2: [1 / 2, 1],
                       4: [3 / 4, 1 / 2, 1],
                       5: [4 / 5, 3 / 5, 1],
                       6: [5 / 6, 1 / 2, 2 / 3, 1],
                       7: [6 / 7, 5 / 7, 4 / 7, 1]}
        allele_freq_conclusions = []

        tested_all_allele_freqs = check_freqs[most_probable_ploidy]
        for af in tested_all_allele_freqs:
            if (af == 0.5):
                likelihood = __stats.halfnorm.pdf(mf_list, loc=af, scale=2 / cov_mean).prod()
            elif (af == 1):
                likelihood = __stats.halfnorm.pdf(2 - mf_list, loc=af, scale=2 / cov_mean).prod()
            else:
                likelihood = __stats.norm.pdf(mf_list, loc=af, scale=2 / cov_mean).prod()
            allele_freq_conclusions.append(likelihood)
        allele_freq_conclusions = __np.array(allele_freq_conclusions)

        if (__np.argmax(allele_freq_conclusions) == 0):
            most_probable_loh = 0
        else:
            most_probable_loh = 1

    return most_probable_ploidy, most_probable_loh



try:
    from isomut2py import HOME
    from isomut2py import SAMTOOLS_MAX_DEPTH

    import sys as __sys
    import time as __time
    import numpy as __np
    import subprocess as __subprocess
    import os as __os
    import multiprocessing as __multiprocessing
    import glob as __glob
except ImportError:
    print('ImportError in isomut2py.processing, ploidy estimation and mutation calling functions might not work.')


def __define_parallel_blocks(genome_length, n_min_block, chromosomes, chrom_length, level=0):
    """

    Calculate blocks of parallelization on the reference genome.

    The number of returned blocks, are higher than min_block_no.
    No block overlap chromosomes, because samtools does not accept region calls overlapping chromosomes.

    :param genome_length: the total length of the genome in basepairs (int)
    :param n_min_block: The approximate number of blocks to partition the analysed genome to for parallel computing. The actual number might be slightly larger that this. (int)
    :param chromosomes: list of chromosomes to analyse (list of str)
    :param chrom_length: list of chromosome lengths (list of int)
    :param level: the level of indentation used in verbose output (default: 0) (int)

    :returns: list of tuples, each tuple is a block: [(chrom, posStart, posEnd), ...]

    """

    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S", __time.localtime()) + ' - Defining parallel blocks ...')
    __sys.stdout.flush()  # jupyter notebook needs this to have constant feedback

    # set maximum block size
    BLOCKSIZE = int(__np.floor(genome_length / n_min_block))

    # calculate blocks
    blocks = []
    # loop over chroms
    for chrom, leng in zip(chromosomes, chrom_length):
        pointer = 0
        # until chrom is chopped into pieces
        while (pointer < leng):
            block_size = min(BLOCKSIZE, leng - pointer)
            blocks.append([chrom, pointer, pointer + block_size])
            pointer += block_size

    # return chr, posmin, posmax of blocks
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S", __time.localtime()) + ' - Done\n')
    __sys.stdout.flush()
    return blocks

def __temp_file_from_block(chrom,
                           from_pos,
                           to_pos,
                           ref_fasta,
                           input_dir,
                           bam_filename,
                           output_dir,
                           windowsize=10000,
                           shiftsize=3000,
                           min_noise=0.1,
                           print_every_nth=1000,
                           base_quality_limit=0,
                           samtools_fullpath='samtools',
                           samtools_flags=None,
                           bedfile = None):
    """

    Run the samtools mpileup command and the PEprep C application in the system shell.

    Creates a temporary file for ploidy estimation by filtering positions based on their reference nucleotide frequency.

    One run is performed on a section of a chromosome.
    Specific parameters are set according to the PloidyEstimation object attributes.
    Automatically logs samtools stderr output to output_dir/samtools.log where output_dir is the output_dir attribute of the PloidyEstimation object.
    Results are saved to a file which name is created from the blocks parameters. (chrom, posStart, posEnd)

    :param chrom: block chromosome (str)
    :param from_pos: block posStart (int)
    :param to_pos:  block posEnd (int)
    :param ref_fasta: path to the reference genome fasta file (str)
    :param input_dir: path to the folder where bam files are located (str)
    :param bam_filename: list of bam filenames to analyse (list of str)
    :param output_dir: path to the directory where results should be saved (str)
    :param windowsize: windowsize for initial coverage smoothing with a moving average method (default: 10000) (int)
    :param shiftsize: shiftsize for initial coverage smoothing with a moving average method (default: 3000) (int)
    :param min_noise: The minimum frequency of non-reference or reference bases detected for a position to be considered for LOH detection. Setting it too small will result in poor noise filtering, setting it too large will result in a decreased number of measurement points. (default: 0.1) (float in range(0,1))
    :param print_every_nth: Even though LOH detection is limited to the positions with a noise level larger that min_noise, ploidy estimation is based on all the genomic positions meeting the above set criteria. By setting the attribute print_every_nth, the number of positions used can be controlled. Setting it large will result in overlooking ploidy variations in shorter genomic ranges, while setting it too small can cause an increase in both memory usage and computation time. Decrease only if a relatively short genome is analysed. (default: 100) (int)
    :param base_quality_limit: The base quality limit used by samtools in order to decide if a base should be included in the pileup file. (default: 0) (int)
    :param samtools_fullpath: The path to samtools on the computer. (default: "samtools") (str)
    :param samtools_flags: The samtools flags to be used for pileup file generation. (default: " -B -d 1000 ") (str)
    :param bedfile: path to the the bedfile to limit ploidy estimation to a specific region of the genome (default: None) (str)

    :returns: status of the run (0 if successful, otherwise 1)

    """

    # build the command
    cmd = ' ' + samtools_fullpath + '  mpileup'
    if samtools_flags != None:
        cmd += ' ' + samtools_flags
    cmd += ' -f ' + ref_fasta
    cmd += ' -r ' + chrom + ':' + str(from_pos) + '-' + str(to_pos) + ' '
    if (bedfile != None):
        cmd += ' -l ' + bedfile + ' '
    cmd += input_dir + '/' + bam_filename + ' '
    cmd += ' 2>> ' + output_dir + '/samtools.log | '+HOME+'/C/PEprep '
    cmd += ' '.join(map(str, [windowsize, shiftsize, min_noise, print_every_nth,
                              base_quality_limit, bam_filename])) + ' '
    cmd += ' > ' + output_dir + '/PEtmp_blockfile_' + chrom + '_' + str(from_pos) + '_' + str(to_pos) + '.csv'

    return __subprocess.check_call(cmd, shell=True)

def PE_prepare_temp_files(ref_fasta,
                            input_dir,
                            bam_filename,
                            output_dir,
                            genome_length,
                            n_min_block,
                            n_conc_blocks,
                            chromosomes,
                            chrom_length,
                            windowsize=10000,
                            shiftsize=3000,
                            min_noise=0.1,
                            print_every_nth=1000,
                            base_quality_limit=0,
                            samtools_fullpath='samtools',
                            samtools_flags=None,
                            bedfile = None,
                            level=0):
    """

    Prepares temporary files for ploidy estimation, by averaging coverage in moving windows for the whole genome and collecting positions with
    reference allele frequencies in the [min_noise, 1-min_noise] range.

    :param ref_fasta: path to the reference genome fasta file (str)
    :param input_dir: path to the folder where bam files are located (str)
    :param bam_filename: list of bam filenames to analyse (list of str)
    :param output_dir: path to the directory where results should be saved (str)
    :param genome_length: the total length of the genome in basepairs (int)
    :param n_min_block: The approximate number of blocks to partition the analysed genome to for parallel computing. The actual number might be slightly larger that this. (default: 200) (int)
    :param n_conc_blocks: The number of blocks to process at the same time. (default: 4) (int)
    :param chromosomes: list of chromosomes in the genome (list of str)
    :param chrom_length: list of chromosome lengths in the genome (list of int)
    :param windowsize: windowsize for initial coverage smoothing with a moving average method (default: 10000) (int)
    :param shiftsize: shiftsize for initial coverage smoothing with a moving average method (default: 3000) (int)
    :param min_noise: The minimum frequency of non-reference or reference bases detected for a position to be considered for LOH detection. Setting it too small will result in poor noise filtering, setting it too large will result in a decreased number of measurement points. (default: 0.1) (float in range(0,1))
    :param print_every_nth: Even though LOH detection is limited to the positions with a noise level larger that min_noise, ploidy estimation is based on all the genomic positions meeting the above set criteria. By setting the attribute print_every_nth, the number of positions used can be controlled. Setting it large will result in overlooking ploidy variations in shorter genomic ranges, while setting it too small can cause an increase in both memory usage and computation time. Decrease only if a relatively short genome is analysed. (default: 100) (int)
    :param base_quality_limit: The base quality limit used by samtools in order to decide if a base should be included in the pileup file. (default: 0) (int)
    :param samtools_fullpath: The path to samtools on the computer. (default: "samtools") (str)
    :param samtools_flags: The samtools flags to be used for pileup file generation. (default: " -B -d 1000 ") (str)
    :param bedfile: path to the the bedfile to limit ploidy estimation to a specific region of the genome (default: None) (str)
    :param level: the level of indentation used in verbose output (default: 0) (int)

    """
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S",
                                         __time.localtime()) + ' - Preparing for parallelization...')

    # define blocks and create args
    blocks = __define_parallel_blocks(genome_length=genome_length,
                                      n_min_block=n_min_block,
                                      chromosomes=chromosomes,
                                      chrom_length=chrom_length,
                                      level=level+1)

    args = []
    for block in blocks:
        args.append([block[0], block[1], block[2], ref_fasta, input_dir, bam_filename, output_dir, windowsize,
                     shiftsize, min_noise, print_every_nth, base_quality_limit, samtools_fullpath, samtools_flags,
                     bedfile])

    # creating output directory (and cleaning if it already exists)
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S",
                                         __time.localtime()) + ' - (All output files will be written to ' + output_dir + ')\n')
    if (__os.path.isdir(output_dir)):
        cmd = 'rm -r ' + output_dir
        __subprocess.call(cmd, shell=True)
    cmd = 'mkdir ' + output_dir
    __subprocess.call(cmd, shell=True)

    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S",
                                         __time.localtime()) + ' - Generating temporary files, number of blocks to run: ' + str(
        len(args)))
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S", __time.localtime()) + ' - Currently running: ', end=' ')
    # start first n concurrent block
    procs = []
    for i in range(n_conc_blocks):
        procs.append(__multiprocessing.Process(target=__temp_file_from_block, args=args[len(procs)]))
        procs[-1].start()
        print(str(len(procs)), end=" ")
        __sys.stdout.flush()

    # when one finished restart the next
    while (len(procs) != len(args)):
        if sum([procs[i].is_alive() for i in range(len(procs))]) < n_conc_blocks and len(procs) != len(args):
            procs.append(__multiprocessing.Process(target=__temp_file_from_block, args=args[len(procs)]))
            procs[-1].start()
            print(str(len(procs)), end=" ")
            __sys.stdout.flush()

        # for i in range(1, n_conc_blocks + 1):
        #     # one finished start another one
        #     if (procs[-i].is_alive() == False and len(procs) != len(args)):
        #         procs.append(__multiprocessing.Process(target=__temp_file_from_block, args=args[len(procs)]))
        #         procs[-1].start()
        #         print(str(len(procs)), end=" ")
        #         __sys.stdout.flush()
        __time.sleep(0.1)

    # wait now only the last ones running
    finished = False
    while (not finished):
        finished = True
        for proc in procs:
            if (proc.is_alive() == True):
                finished = False
        __time.sleep(0.1)

    print('\n')
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S",
                                         __time.localtime()) + ' - Temporary files created, merging, cleaning up...\n')

    # collecting temp position files to single files for each chromosome
    for c in chromosomes:
        cmd = 'cat ' + output_dir + '/PEtmp_blockfile_' + c + '_* > ' + output_dir + '/PEtmp_fullchrom_' + c + '.txt'
        __subprocess.check_call(cmd, shell=True)
        cmd = 'rm ' + output_dir + '/PEtmp_blockfile_' + c + '_*'
        __subprocess.check_call(cmd, shell=True)


def __run_isomut2_on_block(chrom,
                           from_pos,
                           to_pos,
                           ref_fasta,
                           bam_filename,
                           input_dir,
                           output_dir,
                           samtools_fullpath='samtools',
                           samtools_flags = None,
                           bedfile = None,
                           min_sample_freq = 0.21,
                           min_other_ref_freq = 0.95,
                           cov_limit = 5,
                           base_quality_limit = 30,
                           min_gap_dist_snv = 0,
                           min_gap_dist_indel = 20,
                           constant_ploidy = 2,
                           ploidy_info_filepath = 'no_ploidy_info',
                           unique_mutations_only = True,
                           print_shared_by_all = False):
    """

    Run the samtools mpileup command and the isomut2 C application in the system shell.

    Creates a temporary file of potential mutations by filtering positions based on the attributes of the MutationCaller object.

    One run is performed on a section of a chromosome.
    Automatically logs samtools stderr output to output_dir/samtools.log where output_dir is the output_dir attribute of the MutationCaller object.
    Results are saved to a file which name is created from the blocks parameters. (chrom, posStart, posEnd)

    :param chrom: block chromosome (str)
    :param from_pos: block posStart (int)
    :param to_pos:  block posEnd (int)
    :param ref_fasta: path to the reference genome fasta file (str)
    :param bam_filename: A list of the name(s) of the bam file(s) of the sample(s). (Without path, eg. ["sample_1.bam", "sample_2.bam", "sample_3.bam", ...].) (list of str)
    :param input_dir: The path to the directory, where the bam file(s) of the sample(s) is/are located. (str)
    :param output_dir: The path to a directory that can be used for temporary files and output files. The user must have permission to write the directory. (str)
    :param samtools_fullpath: The path to samtools on the computer. (default: "samtools") (str)
    :param samtools_flags: The samtools flags to be used for pileup file generation. (default: " -B -d 1000 ") (str)
    :param bedfile: Path to a bedfile containing a list of genomic regions, if mutations are only sought in a limited part of the genome. (default: None) (str)
    :param min_sample_freq: The minimum frequency of the mutated base at a given position in the mutated sample(s). (default: 0.21) (float)
    :param min_other_ref_freq: The minimum frequency of the reference base at a given position in the non-mutated sample(s). (default: 0.95) (float)
    :param cov_limit: The minimum coverage at a given position in the mutated sample(s). (default: 5) (float)
    :param base_quality_limit: The base quality limit used by samtools in order to decide if a base should be included in the pileup file. (default: 30) (int)
    :param min_gap_dist_snv: Minimum genomic distance from an identified SNV for a position to be considered as a potential mutation. (default: 0) (int)
    :param min_gap_dist_indel: Minimum genomic distance from an identified indel for a position to be considered as a potential mutation. (default: 20) (int)
    :param constant_ploidy: The default ploidy to be used in regions not contained in ploidy_info_filepath. (default: 2) (int)
    :param ploidy_info_filepath: Path to the file containing ploidy information (see below for details) of the samples. (default: None) (str)
    :param unique_mutations_only: If True, only those mutations are sought that are unique to specific samples. Setting it to False greatly increases computation time, but allows for the detection of shared mutations and thus the analysis of phylogenetic connections between samples. If True, print_shared_by_all is ignored. (default: True) (boolean)
    :param print_shared_by_all: If False, mutations that are present in all analysed samples are not printed to the output files. This decreases both memory usage and computation time. (default: False) (boolean)

    :returns: status of the run (0 if successful, otherwise 1)

    """
    # build the command
    cmd = ' ' + samtools_fullpath + '  mpileup ' + samtools_flags
    cmd += ' -f ' + ref_fasta
    cmd += ' -r ' + chrom + ':' + str(from_pos) + '-' + str(to_pos) + ' '
    if bedfile is not None:
        cmd += ' -l ' + bedfile + ' '
    for bam_file in bam_filename:
        cmd += input_dir + bam_file + ' '
    cmd += ' 2>> ' + output_dir + '/samtools.log | '+HOME+'/C/isomut2 '
    cmd += ' '.join(map(str, [min_sample_freq, min_other_ref_freq, cov_limit,
                              base_quality_limit, min_gap_dist_snv, min_gap_dist_indel,
                              constant_ploidy, ploidy_info_filepath, unique_mutations_only, int(print_shared_by_all)]))
    for bam_file in bam_filename:
        cmd += ' ' + __os.path.basename(bam_file)
    cmd += ' > ' + output_dir + '/tmp_isomut2_' + chrom + '_' + str(from_pos) + '_' + str(
        to_pos) + '_mut.csv  '

    return __subprocess.check_call(cmd, shell=True)


def __run_isomut2_in_parallel(genome_length,
                              n_min_block,
                              n_conc_blocks,
                              chromosomes,
                              chrom_length,
                              ref_fasta,
                              bam_filename,
                              input_dir,
                              output_dir,
                              samtools_fullpath='samtools',
                              samtools_flags = None,
                              bedfile = None,
                              min_sample_freq = 0.21,
                              min_other_ref_freq = 0.95,
                              cov_limit = 5,
                              base_quality_limit = 30,
                              min_gap_dist_snv = 0,
                              min_gap_dist_indel = 20,
                              constant_ploidy = 2,
                              ploidy_info_filepath = 'no_ploidy_info',
                              unique_mutations_only = True,
                              print_shared_by_all = False,
                              level=0):
    """

    Generates temporary files of potential mutations, by running the __run_isomut2_on_block() function multiple times in parallel for the whole
    analysed part of the genome.

    :param genome_length: the total length of the genome in basepairs (int)
    :param n_min_block: The approximate number of blocks to partition the analysed genome to for parallel computing. The actual number might be slightly larger that this. (default: 200) (int)
    :param n_conc_blocks: The number of blocks to process at the same time. (default: 4) (int)
    :param chromosomes: list of chromosomes in the genome (list of str)
    :param chrom_length: list of chromosome lengths in the genome (list of int)
    :param ref_fasta: path to the reference genome fasta file (str)
    :param bam_filename: A list of the name(s) of the bam file(s) of the sample(s). (Without path, eg. ["sample_1.bam", "sample_2.bam", "sample_3.bam", ...].) (list of str)
    :param input_dir: The path to the directory, where the bam file(s) of the sample(s) is/are located. (str)
    :param output_dir: The path to a directory that can be used for temporary files and output files. The user must have permission to write the directory. (str)
    :param samtools_fullpath: The path to samtools on the computer. (default: "samtools") (str)
    :param samtools_flags: The samtools flags to be used for pileup file generation. (default: " -B -d 1000 ") (str)
    :param bedfile: Path to a bedfile containing a list of genomic regions, if mutations are only sought in a limited part of the genome. (default: None) (str)
    :param min_sample_freq: The minimum frequency of the mutated base at a given position in the mutated sample(s). (default: 0.21) (float)
    :param min_other_ref_freq: The minimum frequency of the reference base at a given position in the non-mutated sample(s). (default: 0.95) (float)
    :param cov_limit: The minimum coverage at a given position in the mutated sample(s). (default: 5) (float)
    :param base_quality_limit: The base quality limit used by samtools in order to decide if a base should be included in the pileup file. (default: 30) (int)
    :param min_gap_dist_snv: Minimum genomic distance from an identified SNV for a position to be considered as a potential mutation. (default: 0) (int)
    :param min_gap_dist_indel: Minimum genomic distance from an identified indel for a position to be considered as a potential mutation. (default: 20) (int)
    :param constant_ploidy: The default ploidy to be used in regions not contained in ploidy_info_filepath. (default: 2) (int)
    :param ploidy_info_filepath: Path to the file containing ploidy information (see below for details) of the samples. (default: None) (str)
    :param unique_mutations_only: If True, only those mutations are sought that are unique to specific samples. Setting it to False greatly increases computation time, but allows for the detection of shared mutations and thus the analysis of phylogenetic connections between samples. If True, print_shared_by_all is ignored. (default: True) (boolean)
    :param print_shared_by_all: If False, mutations that are present in all analysed samples are not printed to the output files. This decreases both memory usage and computation time. (default: False) (boolean)
    :param level: the level of indentation used in verbose output (default: 0) (int)

    """


    # define blocks and create args
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S",
                                         __time.localtime()) + ' - Preparations for parallelization:\n')

    blocks=__define_parallel_blocks(genome_length=genome_length, n_min_block=n_min_block, chromosomes=chromosomes,
                                    chrom_length=chrom_length, level=level+1)
    args = []
    for block in blocks:
        args.append([block[0], block[1], block[2], ref_fasta, bam_filename, input_dir, output_dir, samtools_fullpath,
                     samtools_flags, bedfile, min_sample_freq, min_other_ref_freq, cov_limit, base_quality_limit,
                     min_gap_dist_snv, min_gap_dist_indel, constant_ploidy, ploidy_info_filepath, unique_mutations_only,
                     print_shared_by_all])

    # creating output directory (and cleaning if it already exists)
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S",
                                         __time.localtime()) + ' - (All output files will be written to ' + output_dir + ')\n')
    if (__os.path.isdir(output_dir)):
        cmd = 'rm -r ' + output_dir
        __subprocess.call(cmd, shell=True)
    cmd = 'mkdir ' + output_dir
    __subprocess.call(cmd, shell=True)

    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S",
                                         __time.localtime()) + ' - Generating temporary files, total number of blocks to run: ' + str(
        len(args)) + '\n')
    print('\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S", __time.localtime()) + ' - Currently running:', end = ' ')

    # start first n concurrent block
    procs = []
    for i in range(n_conc_blocks):
        procs.append(__multiprocessing.Process(target=__run_isomut2_on_block, args=args[len(procs)]))
        procs[-1].start()
        print(str(len(procs)), end=" ")
        __sys.stdout.flush()

    # when one finished restart teh next
    while (len(procs) != len(args)):
        if sum([procs[i].is_alive() for i in range(len(procs))]) < n_conc_blocks and len(procs) != len(args):
            procs.append(__multiprocessing.Process(target=__run_isomut2_on_block, args=args[len(procs)]))
            procs[-1].start()
            print(str(len(procs)), end=" ")
            __sys.stdout.flush()

        # for i in range(1, n_conc_blocks + 1):
        #     # one finished start another one
        #     if (procs[-i].is_alive() == False and len(procs) != len(args)):
        #         procs.append(__multiprocessing.Process(target=__run_isomut2_on_block, args=args[len(procs)]))
        #         procs[-1].start()
        #         print(str(len(procs)), end=" ")
        __time.sleep(0.1)

    # wait now only the last ones running
    finished = False
    while (not finished):
        finished = True
        for proc in procs:
            if (proc.is_alive() == True):
                finished = False
        __time.sleep(0.1)

    print('\n' + '\t' * level + __time.strftime("%Y-%m-%d %H:%M:%S", __time.localtime()) + ' - Finished this round.\n')

def __collect_mutations(output_dir, mut_type, tmp_filename='tmp_isomut2_*_mut.csv'):
    """

    Collects the mutations from multiple temporary files.

    :param output_dir: path to the directory where temporary files are located (str)
    :param mut_type: The type of mutations to collect ("SNV" or "indel"). (str)
    :param tmp_filename: The regular expression for the name of the temporary files to scan through. (default: "tmp_isomut2_*_mut.csv") (str)

    """
    # print header
    header = "#sample_name\tchr\tpos\ttype\tscore\tref\tmut\tcov\tmut_freq\tcleanliness\tploidy\n"
    with open(output_dir + '/all_' + mut_type + 's.isomut2', 'w') as f:
        f.write(header)

    # copy all lines from temporary files (except the header) and sort them by chr, pos
    cmd = 'tail -q -n+2 ' + output_dir + '/' + tmp_filename + ' | '
    if (mut_type == 'indel'):
        cmd += 'awk \'$4=="INS" || $4=="DEL" {print}\' | '
    elif (mut_type == 'SNV'):
        cmd += 'awk \'$4=="SNV" {print}\' | '
    cmd += 'sort -n -k2,2 -k3,3 >> ' + output_dir + '/all_' + mut_type + 's.isomut2'
    __subprocess.check_call(cmd, shell=True)

def __prepare_for_2nd_run(mutcallObject):
    """

    If use_local_realignment = True, isomut2 is run a second time with different samtools settings. This function generates a bedfile from the
    initially identified potential mutations from the first round, which is used to decrease computation time for the second round. Original
    temporary files are saved for later use under a different name.

    :param mutcallObject: an isomut2py.MutationCaller object

    """
    # create bedfile for SNVs for post processing
    __subprocess.check_call(
        'tail -q -n+2 ' + mutcallObject.output_dir + '/tmp_isomut2_*_mut.csv |\
        awk \'$4=="SNV" {print}\' | \
        sort -n -k2,2 -k3,3 | cut -f 2,3 > ' + mutcallObject.output_dir + '/tmp_isomut2.bed', shell=True)

    # renaming original files so they are not overwritten
    for init_res in __glob.glob(mutcallObject.output_dir + '/tmp_isomut2_*_mut.csv'):
        new_file_name = mutcallObject.output_dir + '/' + init_res.split('.')[-2].split('/')[-1] + '_orig.csv'
        __subprocess.check_call('mv ' + init_res + ' ' + new_file_name, shell=True)

    # change parameter values for postprocessing, save originals
    mutcallObject.__base_quality_limit_orig = mutcallObject.base_quality_limit
    mutcallObject.base_quality_limit = 13
    mutcallObject.__min_other_ref_freq_orig = mutcallObject.min_other_ref_freq
    mutcallObject.min_other_ref_freq = 0
    mutcallObject.__samtools_flags_orig = mutcallObject.samtools_flags
    mutcallObject.samtools_flags = ' -d ' + str(SAMTOOLS_MAX_DEPTH) + ' '
    mutcallObject.__bedfile_orig = mutcallObject.bedfile
    mutcallObject.bedfile = mutcallObject.output_dir + '/tmp_isomut2.bed'

def __pair_muts_with_orig_cleanliness(output_dir):
    """

    If use_local_realignment = True, isomut2 is run a second time with different samtools settings. This function pairs the mutations not filtered
    out in the second round with their original cleanliness values, calculated in the first round.

    :param output_dir: path to the directory where temporary files are located (str)

    """

    header = "#sample_name\tchr\tpos\ttype\tscore\tref\tmut\tcov\tmut_freq\tcleanliness\tploidy\n"

    for new_file_name in __glob.glob(output_dir + '/tmp_isomut2_*_mut.csv'):
        with open(new_file_name) as f_new:
            # let's collect original cleanliness into a dict first
            cleanliness_dict = dict()
            old_file_name = output_dir + '/' + new_file_name.split('.')[-2].split('/')[-1] + '_orig.csv'
            with open(old_file_name) as f_old:
                f_old.readline()  # skip header
                for line in f_old:
                    line_list = line.strip().split('\t')
                    if len(line_list) == 11:
                        key = '_'.join(line_list[1:4] + line_list[5:7])
                        cleanliness_dict[key] = [line_list[9], line_list[0]]
            # now it's okay to remove the old file
            __subprocess.check_call('rm ' + old_file_name, shell=True)
            # now we move on to the new files
            f_new.readline()  # skip header
            final_file_name = output_dir + '/' + new_file_name.split('.')[-2].split('/')[-1] + '_final.csv'
            with open(final_file_name, 'w') as f_final:
                f_final.write(header)
                for line in f_new:
                    line_list = line.strip().split('\t')
                    key = '_'.join(line_list[1:4] + line_list[5:7])
                    if key in cleanliness_dict:
                        f_final.write('\t'.join(
                            [cleanliness_dict[key][1]] + line_list[1:9] + [cleanliness_dict[key][0]] + [
                                line_list[10]]) + '\n')

def __del_tmp_files(output_dir):
    """

    Deletes all temporary files from the output_dir of the MutationCaller object. (All files matching the regular expression "tmp_*".)

    :param output_dir: path to the directory where temporary files are located (str)

    """

    __subprocess.check_call('rm ' + output_dir + '/tmp_*', shell=True)

def __cleanup_after_2nd_run(mutcallObject, level=0):
    """

    If use_local_realignment = True, isomut2 is run a second time with different samtools settings. This function resets parameters after the
    second round to their original values, checks the number of discarded mutations in the second round, pairs non-filtered mutations with their
    original cleanliness values, collects identified mutations into a single file and deletes temporary files.

    :param mutcallObject: an isomut2py.MutationCaller object
    :param level: the level of indentation used in verbose output (default: 0) (int)

    """
    # reset parameter values
    mutcallObject.base_quality_limit = mutcallObject.__base_quality_limit_orig
    del mutcallObject.__base_quality_limit_orig
    mutcallObject.min_other_ref_freq = mutcallObject.__min_other_ref_freq_orig
    del mutcallObject.__min_other_ref_freq_orig
    mutcallObject.samtools_flags = mutcallObject.__samtools_flags_orig
    del mutcallObject.__samtools_flags_orig
    mutcallObject.bedfile = mutcallObject.__bedfile_orig
    del mutcallObject.__bedfile_orig

    print('\t' * (level) + __time.strftime("%Y-%m-%d %H:%M:%S", __time.localtime()) + ' - Finalizing output...')

    # check number of discarded mutations
    number_of_lines_old = __subprocess.check_output(
        'cat ' + mutcallObject.output_dir + '/tmp_isomut2_*_mut_orig.csv | grep -v "^#" | wc -l', shell=True)
    number_of_lines_old = int(number_of_lines_old.strip())
    number_of_lines_new = __subprocess.check_output(
        'cat ' + mutcallObject.output_dir + '/tmp_isomut2_*_mut.csv | grep -v "^#" | wc -l', shell=True)
    number_of_lines_new = int(number_of_lines_new.strip())
    print('\t' * (level + 1) + ' - Discarded mutations due to BAQ realignment: ' + str(
        round(1 - (number_of_lines_new / number_of_lines_old), 2))) + '% (' + str(
        number_of_lines_old - number_of_lines_new) + ' total mutations).'

    # pair mutations with their old cleanliness values
    __pair_muts_with_orig_cleanliness(mutcallObject.output_dir)

    # now we collect SNVs
    __collect_mutations(output_dir=mutcallObject.output_dir, mut_type='SNV', tmp_filename='tmp_isomut2_*_mut_final.csv')

    # deleting temporary files
    print('\t' * (level) + __time.strftime("%Y-%m-%d %H:%M:%S",
                                           __time.localtime()) + ' - Cleaning up temporary files...')

    __del_tmp_files(mutcallObject.output_dir)

def single_run(mutcallObject,level=0):
    """

    Runs the mutation detection pipeline on the MutationCaller object when no local realignment is used during the process.

    :param mutcallObject: an isomut2py.MutationCaller object
    :param level: the level of indentation used in verbose output (default: 0) (int)

    """
    print('\n' + '\t' * (level + 1) + __time.strftime("%Y-%m-%d %H:%M:%S",
                                                      __time.localtime()) + ' - Running IsoMut2 without local realignment: \n\n')
    __run_isomut2_in_parallel(genome_length=mutcallObject.genome_length,
                              n_min_block=mutcallObject.n_min_block,
                              n_conc_blocks=mutcallObject.n_conc_blocks,
                              chromosomes=mutcallObject.chromosomes,
                              chrom_length=mutcallObject.chrom_length,
                              ref_fasta=mutcallObject.ref_fasta,
                              bam_filename=mutcallObject.bam_filename,
                              input_dir=mutcallObject.input_dir,
                              output_dir=mutcallObject.output_dir,
                              samtools_fullpath=mutcallObject.samtools_fullpath,
                              samtools_flags = mutcallObject.samtools_flags,
                              bedfile = mutcallObject.bedfile,
                              min_sample_freq = mutcallObject.min_sample_freq,
                              min_other_ref_freq = mutcallObject.min_other_ref_freq,
                              cov_limit = mutcallObject.cov_limit,
                              base_quality_limit = mutcallObject.base_quality_limit,
                              min_gap_dist_snv = mutcallObject.min_gap_dist_snv,
                              min_gap_dist_indel = mutcallObject.min_gap_dist_indel,
                              constant_ploidy = mutcallObject.constant_ploidy,
                              ploidy_info_filepath = mutcallObject.ploidy_info_filepath,
                              unique_mutations_only = mutcallObject.unique_mutations_only,
                              print_shared_by_all = mutcallObject.print_shared_by_all,
                              level=level+2)

    print('\t' * (level + 2) + __time.strftime("%Y-%m-%d %H:%M:%S", __time.localtime()) + ' - Finalizing output...')

    __collect_mutations(output_dir=mutcallObject.output_dir, mut_type='indel')
    __collect_mutations(output_dir=mutcallObject.output_dir, mut_type='SNV')

    print('\t' * (level + 2) + __time.strftime("%Y-%m-%d %H:%M:%S",
                                               __time.localtime()) + ' - Cleaning up temporary files...')
    #         subprocess.check_call('rm '+self.output_dir+'/tmp_isomut2_*_mut.csv',shell=True)
    __del_tmp_files(output_dir=mutcallObject.output_dir)


def double_run(mutcallObject, level=0):
    """

    Runs the mutation detection pipeline on the MutationCaller object when with local realignment.

    :param mutcallObject: an isomut2py.MutationCaller object
    :param level: the level of indentation used in verbose output (default: 0) (int)

    """

    # run 1st
    print('\n' + '\t' * (level + 1) + __time.strftime("%Y-%m-%d %H:%M:%S",
                                                      __time.localtime()) + ' - Running IsoMut2: round 1/2 \n\n')
    __run_isomut2_in_parallel(genome_length=mutcallObject.genome_length,
                              n_min_block=mutcallObject.n_min_block,
                              n_conc_blocks=mutcallObject.n_conc_blocks,
                              chromosomes=mutcallObject.chromosomes,
                              chrom_length=mutcallObject.chrom_length,
                              ref_fasta=mutcallObject.ref_fasta,
                              bam_filename=mutcallObject.bam_filename,
                              input_dir=mutcallObject.input_dir,
                              output_dir=mutcallObject.output_dir,
                              samtools_fullpath=mutcallObject.samtools_fullpath,
                              samtools_flags = mutcallObject.samtools_flags,
                              bedfile = mutcallObject.bedfile,
                              min_sample_freq = mutcallObject.min_sample_freq,
                              min_other_ref_freq = mutcallObject.min_other_ref_freq,
                              cov_limit = mutcallObject.cov_limit,
                              base_quality_limit = mutcallObject.base_quality_limit,
                              min_gap_dist_snv = mutcallObject.min_gap_dist_snv,
                              min_gap_dist_indel = mutcallObject.min_gap_dist_indel,
                              constant_ploidy = mutcallObject.constant_ploidy,
                              ploidy_info_filepath = mutcallObject.ploidy_info_filepath,
                              unique_mutations_only = mutcallObject.unique_mutations_only,
                              print_shared_by_all = mutcallObject.print_shared_by_all,
                              level=level+2)

    # prepare for 2nd round
    print('\t' * (level + 2) + __time.strftime("%Y-%m-%d %H:%M:%S",
                                               __time.localtime()) + ' - Preparing files for post-processing...')
    __collect_mutations(output_dir=mutcallObject.output_dir, mut_type='indel')
    __prepare_for_2nd_run(mutcallObject)

    # run 2nd
    print('\t' * (level + 1) + __time.strftime("%Y-%m-%d %H:%M:%S",
                                               __time.localtime()) + ' - Running IsoMut2: round 2/2\n\n')
    __run_isomut2_in_parallel(genome_length=mutcallObject.genome_length,
                              n_min_block=mutcallObject.n_min_block,
                              n_conc_blocks=mutcallObject.n_conc_blocks,
                              chromosomes=mutcallObject.chromosomes,
                              chrom_length=mutcallObject.chrom_length,
                              ref_fasta=mutcallObject.ref_fasta,
                              bam_filename=mutcallObject.bam_filename,
                              input_dir=mutcallObject.input_dir,
                              output_dir=mutcallObject.output_dir,
                              samtools_fullpath=mutcallObject.samtools_fullpath,
                              samtools_flags=mutcallObject.samtools_flags,
                              bedfile=mutcallObject.bedfile,
                              min_sample_freq=mutcallObject.min_sample_freq,
                              min_other_ref_freq=mutcallObject.min_other_ref_freq,
                              cov_limit=mutcallObject.cov_limit,
                              base_quality_limit=mutcallObject.base_quality_limit,
                              min_gap_dist_snv=mutcallObject.min_gap_dist_snv,
                              min_gap_dist_indel=mutcallObject.min_gap_dist_indel,
                              constant_ploidy=mutcallObject.constant_ploidy,
                              ploidy_info_filepath=mutcallObject.ploidy_info_filepath,
                              unique_mutations_only=mutcallObject.unique_mutations_only,
                              print_shared_by_all=mutcallObject.print_shared_by_all,
                              level=level + 2)

    # cleanup
    __cleanup_after_2nd_run(mutcallObject, level=level + 2)


try:
    from isomut2py import process
    from isomut2py import io
    from isomut2py import bayesian
    from isomut2py import format
    from isomut2py import plot
    from isomut2py import compare

    from datetime import datetime as datetime
    import time as time
    import pandas as pd
    import os as os
    import subprocess as subprocess
    import sys as sys
except ImportError:
    print('ImportError in isomut2py.ploidyestimation, PloidyEstimator object cannot be defined.')

SAMTOOLS_MAX_DEPTH=1000

class PloidyEstimator:
    '''

    The PloidyEstimator class is designed to keep all parameter values, directories and filepaths in one place that are needed for the ploidy
    analysis of a single sample.

    - List of basic parameters:
        - ref_fasta: The path to the fasta file of the reference genome. (str)

        - output_dir: The path to a directory that can be used for temporary files and output files. The user must have permission to write the directory. (str)

        - input_dir: The path to the directory, where the bam file(s) of the sample(s) is/are located. (str)

        - bam_filename: The name of the bam file of the sample. (Without path, eg. "sample_1.bam".) (str)

        - samtools_fullpath: The path to samtools on the computer. (default: "samtools") (str)

    - Other parameters with default values:
        - n_min_block: The approximate number of blocks to partition the analysed genome to for  parallel computing. The actual number might be slightly larger that this. (default: 200) (int)

        - n_conc_blocks: The number of blocks to process at the same time. (default: 4) (int)

        - chromosomes: The list of chromosomes to analyse. (default: all chromosomes included included in the reference genome specified in ref_fasta) (list of str)

        - windowsize: The windowsize used for initial coverage smoothing of the bam file with a moving average method. Setting it too large might disguise CNV effects. (default: 10000) (int)

        - shiftsize: The shiftsize used for the moving average method of the initial coverage smoothing procedure. MUST be smaller than windowsize. (default: 3000) (int)

        - min_noise: The minimum frequency of non-reference or reference bases detected for a position to be considered for LOH detection. Setting it too small will result in poor noise filtering, setting it too large will result in a decreased number of measurement points. (default: 0.1) (float in range(0,1))

        - base_quality_limit: The base quality limit used by samtools in order to decide if a base should be included in the pileup file. (default: 0) (int)

        - print_every_nth: Even though LOH detection is limited to the positions with a noise level larger that min_noise, ploidy estimation is based on all the genomic positions meeting the above set criteria. By setting the attribute print_every_nth, the number of positions used can be controlled. Setting it large will result in overlooking ploidy variations in shorter genomic ranges, while setting it too small can cause an increase in both memory usage and computation time. Decrease only if a relatively short genome is analysed. (default: 100) (int)

        - windowsize_PE: The windowsize used for actual ploidy estimation after the initial coverage smoothing. (default: 1000000) (int)

        - shiftsize_PE: The shiftsize used for actual ploidy estimation after the initial coverage smoothing. MUST be smaller than windowsize_PE. (default: 50000) (int)

        - cov_max: The maximum coverage in a genomic position that can be considered for ploidy estimation. Alignment errors might cause certain genomic positions to have an enormous coverage. These outliers are ignored, when cov_max is set. The value must be set in agreement with the average sequencing depth. Using a low value for a deeply sequenced sample can result in a decreased number of positions to be analysed. (default: 200) (int)

        - cov_min: The minimum coverage in a genomic position that can be considered for ploidy estimation. Sequencing noise might cause certain genomic positions to have very low coverage, frequently merely from misaligned reads. These outliers are ignored, when cov_min is set. The value must be set in agreement with the average sequencing depth. Using a high value for a shallowly sequenced sample can result in a decreased number of positions to be analysed. (default: 5) (int)

        - hc_percentile: The haploid coverage of the sequenced sample is estimated multiple times by fitting a mixture model to the coverage distribution of the sample. The actual value is chosen as a statistical measure of these multiple results, set by the value of hc_percentile. For example, setting hc_percentile to 50 results in using the median of the results. For more details on the suggested values, see Pipek et al. 2018. (default: 75) (int)

        - compare_to_bed: The path to a bed file to compare ploidy estimation results to. (default: None) (str)

        - samtools_flags: The samtools flags to be used for pileup file generation. (default: " -B -d 1000 ") (str)

        - user_defined_hapcov: During ploidy estimation, the haploid coverage is estimated from the coverage distribution of the sample. In some cases, the estimation might not find the real value of the haploid coverage. In these situations, supplying an estimate of the haploid coverage manually might improve the overall ploidy estimation results. If you have a generally diploid genome, using the half of the average coverage can be a good starting point. If user_defined_hapcov is set, hc_percentile is ignored. (default: None) (float)

    '''
    def __init__(self, **kwargs):
        """

        Initializes the MutationCaller object by setting attributes to the ones defined in kwargs.

        :param kwargs: keyword arguments

        """
        for k, v in kwargs.items():
            setattr(self, k, v)


    def __check_params(self, list_of_params):
        """

        Checks and updates parameters values listed in list_of_param.

        :param list_of_params: parameter values to check (array-like)

        """

        def get_default_chroms_with_len(ref_genome, chrom_list):
            """

            Gets the length of chromosomes defined in chrom_list or if chrom_list is empty, gets the default list of chromosomes from a specified
            reference genome fasta file (set in ref_genome) with their length.

            :param ref_genome: The path to the reference genome fasta file.
            :param chrom_list: A predefined list of chromosomes to be included.

            :returns: (chrom, lens):

                - chrom: a list of detected chromosomes (list of str)
                - len: a list of their lengths (list of int)

            """
            newchroms, newlens = [], []
            with open(ref_genome + '.fai') as f_h:
                for line in f_h:
                    chrom, leng = line.split('\t')[0], line.split('\t')[1]
                    if (chrom in chrom_list or len(chrom_list) == 0):
                        newchroms.append(chrom)
                        newlens.append(int(leng))
            if len(chrom_list)>0:
                return chrom_list, [newlens[newchroms.index(c)] for c in chrom_list]
            else:
                return sorted(newchroms), [newlens[newchroms.index(c)] for c in sorted(newchroms)]

        defaults = dict()
        defaults['n_min_block'] = 200
        defaults['n_conc_blocks'] = 4
        defaults['chromosomes'] = None
        defaults['windowsize'] = 10000
        defaults['shiftsize'] = 3000
        defaults['min_noise'] = 0.1
        defaults['base_quality_limit'] = 0
        defaults['windowsize_PE'] = 1000000
        defaults['shiftsize_PE'] = 50000
        defaults['cov_max'] = 200
        defaults['cov_min'] = 5
        defaults['hc_percentile'] = 75
        defaults['print_every_nth'] = 100
        defaults['compare_to_bed'] = 'None'
        defaults['bedfile'] = None
        defaults['samtools_flags'] = ' -B -d ' + str(SAMTOOLS_MAX_DEPTH) + ' '
        defaults['samtools_fullpath'] = 'samtools'
        defaults['user_defined_hapcov'] = None
        defaults['ownbed_filepath'] = None
        defaults['bed_dataframe'] = None
        defaults['coverage_sample'] = None
        defaults['distribution_dict'] = None

        for p in list_of_params:
            if p in ['output_dir', 'input_dir', 'bam_filename']:
                if not hasattr(self, p):
                    raise ValueError('Error: Value of "' + p + '" is not set, cannot proceed.')
                elif (p == 'input_dir' or p == 'bam_filename'):
                    if not os.path.isdir(self.input_dir):
                        raise ValueError(
                            'Error: Input directory "' + self.input_dir + '" does not exist, cannot proceed.')
                    if not os.path.isfile(self.input_dir + '/' + self.bam_filename):
                        raise ValueError(
                            'Error: Bam file "' + self.input_dir + '/' + self.bam_filename + '" does not exist, '
                                                                                             'cannot proceed.')
            elif p == 'samtools_fullpath':
                if not hasattr(self, p):
                    setattr(self, p, 'samtools')
                s = subprocess.getoutput(self.samtools_fullpath).strip().split(' ')[0] == 'Program:'
                if not s:
                    error_msg = 'Error: samtools not found, cannot proceed.'
                    error_msg += '\n'
                    error_msg += 'Specify path to samtools with the "samtools_fullpath" keyword argument.'
                    raise ValueError(error_msg)
            elif p == 'ref_fasta':
                if not hasattr(self, p):
                    raise ValueError('Error: Value of "' + p + '" is not set, cannot proceed.')
                if not os.path.isfile(self.ref_fasta):
                    raise ValueError(
                        'Error: Reference genome file "' + self.ref_fasta + '" does not exist, cannot proceed.')
                if not os.path.isfile(self.ref_fasta + '.fai'):
                    error_msg = 'Error: No faidx file found for reference genome file "' + self.ref_fasta + '", cannot proceed.'
                    error_msg += '\n'
                    error_msg += 'Use the samtools command: samtools faidx [ref.fasta]'
                    raise ValueError(error_msg)
            elif p == 'chromosomes' and (not hasattr(self, 'chromosomes') or not self.chromosomes):
                if not os.path.isfile(self.ref_fasta):
                    raise ValueError(
                        'Error: Reference genome file "' + self.ref_fasta + '" does not exist, cannot proceed.')
                if not os.path.isfile(self.ref_fasta + '.fai'):
                    error_msg = 'Error: No faidx file found for reference genome file "' + self.ref_fasta + '", cannot proceed.'
                    error_msg += '\n'
                    error_msg += 'Use the samtools command: samtools faidx [ref.fasta]'
                    raise ValueError(error_msg)
                c, l = get_default_chroms_with_len(self.ref_fasta, [])
                setattr(self, 'chromosomes', c)
                setattr(self, 'chrom_length', l)
                setattr(self, 'genome_length', sum(l))
            elif p == 'chrom_length' and not hasattr(self, 'chrom_length'):
                if not os.path.isfile(self.ref_fasta):
                    raise ValueError(
                        'Error: Reference genome file "' + self.ref_fasta + '" does not exist, cannot proceed.')
                if not os.path.isfile(self.ref_fasta + '.fai'):
                    error_msg = 'Error: No faidx file found for reference genome file "' + self.ref_fasta + '", cannot proceed.'
                    error_msg += '\n'
                    error_msg += 'Use the samtools command: samtools faidx [ref.fasta]'
                    raise ValueError(error_msg)
                c, l = get_default_chroms_with_len(self.ref_fasta, [])
                setattr(self, 'chromosomes', c)
                setattr(self, 'chrom_length', l)
                setattr(self, 'genome_length', sum(l))
            elif p == 'genome_length':
                setattr(self, p, sum(self.chrom_length))
            elif p == 'estimated_hapcov' and not hasattr(self, 'estimated_hapcov'):
                raise ValueError('Error: Value of "' + p + '" is not set, cannot proceed.')
            elif not hasattr(self, p):
                setattr(self, p, defaults[p])

        if hasattr(self, 'ownbed_filepath') and self.ownbed_filepath is None:
            self.ownbed_filepath = self.output_dir+'/'+self.bam_filename.split('.bam')[0] + '_ploidy.bed'


    def get_coverage_distribution(self, **kwargs):
        """

        Sets the coverage_sample attribute of the PloidyEstimation object to the coverage distribution obtained from the temporary files created by __PE_prepare_temp_files().
        Positions are filtered according to the attributes of the PloidyEstimation object. The number of positions in the final sample is decreased to 2000 for faster inference.

        :param kwargs: keyword arguments for PloidyEstimator object

        :returns: A 2000-element sample of the coverage distribution.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['chromosomes', 'output_dir', 'cov_max', 'cov_min'])

        covs_few = io.get_coverage_distribution(chromosomes=self.chromosomes,
                                                output_dir=self.output_dir,
                                                cov_max=self.cov_max,
                                                cov_min = self.cov_min)

        self.coverage_sample = covs_few
        return covs_few

    def estimate_hapcov_infmix(self, level=0, **kwargs):
        """

        Estimates the haploid coverage of the sample from the appropriate attributes of the PloidyEstimation object.
        If the user_defined_hapcov attribute is set manually, it sets the value of estimated_hapcov to that. Otherwise, a many-component (20)
        Gaussian mixture model is fitted to the coverage histogram of the sample 10 times. Each time, the haploid coverage is estimated from the
        center of the component with the maximal weight in the model. The final estimate of the haploid coverage is calculated as the qth
        percentile of the 10 measurements, with q = hc_percentile. Sets the "estimated_hapcov" attribute to the calculated haploid coverage and the "coverage_sample" attribute to a 2000-element sample of the coverage distribution.

        :param level: the level of indentation used in verbose output (default: 0) (int)
        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['user_defined_hapcov', 'cov_min', 'hc_percentile', 'chromosomes', 'output_dir', 'cov_max'])

        hc, cov_sample = bayesian.estimate_hapcov_infmix(level=level,
                                                         user_defined_hapcov=self.user_defined_hapcov,
                                                         cov_min = self.cov_min,
                                                         hc_percentile = self.hc_percentile,
                                                         chromosomes=self.chromosomes,
                                                         output_dir=self.output_dir,
                                                         cov_max=self.cov_max)

        self.estimated_hapcov = hc
        self.coverage_sample = cov_sample

    def fit_gaussians(self, level=0, **kwargs):
        """

        Fits a 7-component Gaussian mixture model to the coverage distribution of the sample, using the appropriate attributes of the
        PloidyEstimation object. The center of the first Gaussian is initialized from a narrow region around the value of the estimated_hapcov
        attribute. The centers of the other Gaussians are initialized in a region around the value of estimated_hapcov multiplied by consecutive
        whole numbers.

        The parameters of the fitted model (center, sigma and weight) for all seven Gaussians are both saved to the GaussDistParams.pkl file (in
        output_dir, for later reuse) and set as the value of the distribution_dict attribute.

        :param level: the level of indentation used in verbose output (default: 0) (int)
        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        if hasattr(self, 'user_defined_hapcov') and self.user_defined_hapcov is not None:
            self.estimated_hapcov = self.user_defined_hapcov

        self.__check_params(['estimated_hapcov', 'output_dir', 'chromosomes', 'cov_max', 'cov_min'])


        prior_dict = bayesian.fit_gaussians(level=level,
                                            estimated_hapcov=self.estimated_hapcov,
                                            output_dir=self.output_dir,
                                            chromosomes=self.chromosomes,
                                            cov_max=self.cov_max,
                                            cov_min=self.cov_min,
                                            cov_sample=self.coverage_sample)
        self.distribution_dict = prior_dict

    def PE_on_chrom(self, chrom, **kwargs):
        """

        Runs the whole ploidy estimation pipeline on a given chromosome, using the appropriate attributes of the PloidyEstimation object by running PE_on_range() multiple times.
        Prints the results to the file: [self.output_dir]/PE_fullchrom_[chrom].txt.

        :param chrom: the name of the chromosome to analyse (str)
        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['output_dir', 'windowsize_PE', 'shiftsize_PE', 'distribution_dict'])

        bayesian.PE_on_chrom(chrom=chrom,
                             output_dir=self.output_dir,
                             windowsize_PE=self.windowsize_PE,
                             shiftsize_PE=self.shiftsize_PE,
                             distribution_dict=self.distribution_dict)

    def PE_on_whole_genome(self, **kwargs):
        """

        Runs the whole ploidy estimation pipeline on the whole genome by running PE_on_chrom() on all chromosomes.

        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['chromosomes'])

        for c in self.chromosomes:
            print(c, end=" ")
            sys.stdout.flush()
            self.PE_on_chrom(chrom=c)


    def plot_karyotype_summary(self, **kwargs):
        """

        Plots a simple karyotype summary for the whole genome. (Details coming soon.)

        :param kwargs: keyword arguments for PloidyEstimator object

        :returns: a matplotlib figure of the plot

        """

        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

        self.__check_params(['output_dir', 'chromosomes', 'ownbed_filepath', 'cov_min', 'cov_max', 'chrom_length'])

        if not hasattr(self, 'distribution_dict'):
            self.load_cov_distribution_parameters_from_file(filename=kwargs.get("distribution_file"))

        return plot.plot_karyotype_summary(haploid_coverage=self.distribution_dict['mu'][0],
                                           chromosomes=self.chromosomes,
                                           chrom_length=self.chrom_length,
                                           output_dir=self.output_dir,
                                           bed_filename=self.ownbed_filepath,
                                           bed_file_sep=kwargs.get('bed_file_sep', ','),
                                           binsize=kwargs.get('binsize', 1000000),
                                           overlap=kwargs.get('overlap', 50000),
                                           cov_min=self.cov_min,
                                           cov_max=self.cov_max,
                                           min_PL_length=kwargs.get('min_PL_length', 3000000),
                                           chroms_with_text=kwargs.get('chroms_with_text'))


    def get_bed_format_for_sample(self, **kwargs):
        """

        Creates bed file of constant ploidies for a given sample from a file of positional ploidy data. If the ownbed_filepath attribute of the
        PloidyEstimation object is set, saves the bedfile to the path specified there. Otherwise, saves it to the output_dir with the "_ploidy.bed" suffix.
        Also sets the bed_dataframe attribute to the pandas.Dataframe containing the bed file.

        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['chromosomes', 'chrom_length', 'output_dir', 'bam_filename', 'ownbed_filepath'])

        obf, df = format.get_bed_format_for_sample(chromosomes=self.chromosomes,
                                         chrom_length=self.chrom_length,
                                         output_dir=self.output_dir,
                                         bam_filename=self.bam_filename,
                                         ownbed_filepath=self.ownbed_filepath)

        self.ownbed_filepath = obf
        self.bed_dataframe = df

    def plot_karyotype_for_all_chroms(self, return_string=False, **kwargs):
        """

        Plots karyotype information (coverage, estimated ploidy, estimated LOH, reference base frequencies) about the sample for all analysed
        chromosomes.

        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: True) (bool)
        :param kwargs: keyword arguments for PloidyEstimator object

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['chromosomes', 'output_dir'])

        return plot.plot_karyotype_for_all_chroms(chromosomes=self.chromosomes,
                                           output_dir=self.output_dir,
                                           return_string=return_string)

    def generate_HTML_report_for_ploidy_est(self, **kwargs):
        """

        Generates a HTML file with figures displaying the results of ploidy estimation and saves it to output_dir/PEreport.html.

        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['chromosomes', 'output_dir', 'min_noise'])

        plot.generate_HTML_report_for_ploidy_est(chromosomes=self.chromosomes,
                                                 output_dir=self.output_dir,
                                                 min_noise=self.min_noise)

    def load_cov_distribution_parameters_from_file(self, filename=None, **kwargs):
        """

        Loads the parameters of the seven fitted Gaussians to the coverage distribution of the sample from the specified filename (that was saved
        with pickle beforehand). If one such file is available, the computationally expensive ploidy estimation process can be skipped. The
        parameter values will be stored in the attribute "distribution_dict" as a dictionary.

        :param filename: The path to the file with the coverage distribution parameters. (default: [output_dir]/GaussDistParams.pkl) (str)
        :param kwargs: keyword arguments for PloidyEstimator object

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['output_dir'])

        d = io.load_cov_distribution_parameters_from_file(filename=filename, output_dir=self.output_dir)
        self.distribution_dict = d

    def load_bedfile_from_file(self, filename=None, **kwargs):
        """

        Loads the bedfile containing previous ploidy estimated for the given sample from the path specified in filename. The dataframe will be
        stored in the "bed_dataframe" attribute.

        :param filename: The path to the bedfile. (default: [output_dir]/[bam_filename]_ploidy.bed) (str)
        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['output_dir', 'bam_filename'])

        self.bed_dataframe = io.load_bedfile_from_file(filename=filename,
                                                       output_dir=self.output_dir,
                                                       bam_filename=self.bam_filename)

    def run_ploidy_estimation(self, **kwargs):
        """

        Runs the whole ploidy estimation pipeline on the PloidyEstimation object.

        :param kwargs: keyword arguments for PloidyEstimator object

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'ref_fasta', 'input_dir', 'output_dir', 'genome_length',
                             'n_min_block', 'n_conc_blocks', 'chromosomes', 'chrom_length', 'windowsize', 'shiftsize',
                             'min_noise', 'print_every_nth', 'base_quality_limit', 'samtools_fullpath', 'samtools_flags',
                             'bedfile', 'user_defined_hapcov'])

        # running ploidy estimation
        level = 0

        starting_time = datetime.now()
        print('\t' * level + time.strftime("%Y-%m-%d %H:%M:%S",
                                           time.localtime()) + ' - Ploidy estimation for file ' + self.bam_filename)
        print('\n')

        process.PE_prepare_temp_files(ref_fasta=self.ref_fasta,
                                        input_dir=self.input_dir,
                                        bam_filename=self.bam_filename,
                                        output_dir=self.output_dir,
                                        genome_length=self.genome_length,
                                        n_min_block=self.n_min_block,
                                        n_conc_blocks=self.n_conc_blocks,
                                        chromosomes=self.chromosomes,
                                        chrom_length=self.chrom_length,
                                        windowsize=self.windowsize,
                                        shiftsize=self.shiftsize,
                                        min_noise=self.min_noise,
                                        print_every_nth=self.print_every_nth,
                                        base_quality_limit=self.base_quality_limit,
                                        samtools_fullpath=self.samtools_fullpath,
                                        samtools_flags=self.samtools_flags,
                                        bedfile = self.bedfile,
                                        level=level+1)

        if (self.user_defined_hapcov):
            print('\t' * (level + 1) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                     time.localtime()) + ' - Collecting data for coverage distribution, using user-defined haploid coverage (' + str(
                self.user_defined_hapcov) + ')...\n')
            self.estimate_hapcov_infmix(level=level + 2)
        else:
            print('\t' * (level + 1) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                     time.localtime()) + ' - Estimating haploid coverage by fitting an infinite mixture model to the coverage distribution...\n')
            self.estimate_hapcov_infmix(level=level + 2)
            print('\t' * (level + 2) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                     time.localtime()) + ' - Raw estimate for the haploid coverage: ' + str(
                self.estimated_hapcov) + '\n')

        print('\t' * (level + 1) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                 time.localtime()) + ' - Fitting equidistant Gaussians to the coverage distribution using the raw haploid coverage as prior...\n')
        self.fit_gaussians(level=level + 2)
        print('\t' * (level + 2) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                 time.localtime()) + ' - Parameters of the distribution are saved to: ' + self.output_dir + '/GaussDistParams.pkl' + '\n')
        print('\t' * (level + 2) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                 time.localtime()) + ' - Final estimate for the haploid coverage: ' + str(
            self.distribution_dict['mu'][0]) + '\n')
        print('\t' * (level + 1) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                 time.localtime()) + ' - Estimating local ploidy using the previously determined Gaussians as priors on chromosomes: ', end=' ')
        for c in self.chromosomes:
            print(c, end=" ")
            sys.stdout.flush()
            self.PE_on_chrom(chrom=c)

        print('\n\n' + '\t' * (level + 1) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                          time.localtime()) + ' - Generating final bed file... \n')
        self.get_bed_format_for_sample()

        print('\n' + '\t' * (level + 1) + time.strftime("%Y-%m-%d %H:%M:%S",
                                                        time.localtime()) + ' - Generating HTML report... \n')
        self.generate_HTML_report_for_ploidy_est()

        finish_time = datetime.now()
        total_time = finish_time - starting_time
        total_time_h = int(total_time.seconds / 3600)
        total_time_m = int((total_time.seconds % 3600) / 60)
        total_time_s = (total_time.seconds % 3600) % 60
        print('\n' + '\t' * level + time.strftime("%Y-%m-%d %H:%M:%S",
                                                  time.localtime()) + ' - Ploidy estimation finished. (' + str(
            total_time.days) + ' day(s), ' + str(total_time_h) + ' hour(s), ' + str(total_time_m) + ' min(s), ' + str(
            total_time_s) + ' sec(s).)')



    def compare_with_other(self, other, minLen=2000, minQual=0.1):
        """

        Compare ploidy estimation results with another PloidyEstimation object or a bed file.

        :param other: The other PloidyEstimation object or the path to the other bedfile. (isomut2py.PloidyEstimation or str)
        :param minLen: The minimum length of a region to be considered different from the other object or file. (int)
        :param minQual: The minimum quality of a region to be considered different from the other object or file. (float)

        """
        if other.__class__ == self.__class__:
            compare.compare_with_other_PloidyEstimator(self, other, minLen, minQual)
        elif other.__class__ == str or other.__class__ == pd.core.frame.DataFrame:
            compare.compare_with_bed(bed_dataframe=self.bed_dataframe, other=other, minLen=minLen)
        else:
            error_msg = 'Error: For ploidy estimation comparison, the type of the other object must be either str or PloidyEstimation.'
            raise ValueError(error_msg)

    def plot_coverage_distribution(self, **kwargs):
        """

        Plot the coverage distribution of the sample.

        :param kwargs: keyword arguments for PloidyEstimator object

        :returns: a matplotlib figure of the coverage distribution

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['coverage_sample', 'distribution_dict', 'chromosomes', 'output_dir',
                             'cov_max', 'cov_min'])

        return plot.plot_coverage_distribution(cov_sample = self.coverage_sample,
                                               chromosomes=self.chromosomes,
                                               output_dir = self.output_dir,
                                               cov_max = self.cov_max,
                                               cov_min = self.cov_min,
                                               distribution_dict=self.distribution_dict)


# command line parameter "-d" for samtools, used for all pileup file generation
SAMTOOLS_MAX_DEPTH=1000
# path to package
HOME = __path__[0]

__version__ = '2.0.5'

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=DeprecationWarning)

from isomut2py import mutationcalling
from isomut2py import ploidyestimation
from isomut2py import bayesian
from isomut2py import compare
from isomut2py import format
from isomut2py import io
from isomut2py import plot
from isomut2py import postprocess
from isomut2py import process
from isomut2py import examples

import subprocess as __subprocess
import os as __os

# warnings.resetwarnings()

##### this whole compiling thing should be in the setup script
##### https://stackoverflow.com/questions/20288711/post-install-script-with-python-setuptools

# compiling C scripts
for __fname in __os.listdir(HOME+'/C'):
    if __fname.split('.')[-1] != 'h' and __fname.split('.')[-1] != 'c':
        __subprocess.call('rm ' + HOME + '/C/' + __fname, shell=True)

print('Compiling C scripts...')
cwd = __os.getcwd()
__compiled = []
__os.chdir(HOME + '/C')
__cmd = 'gcc -c -O3 isomut2_lib.c fisher.c  -W -Wall'
__compiled.append(__subprocess.call(__cmd, shell=True))
__cmd = 'gcc -O3 -o isomut2 isomut2.c isomut2_lib.o fisher.o -lm -W -Wall'
__compiled.append(__subprocess.call(__cmd, shell=True))
__cmd = 'gcc -O3 -o PEprep PEprep.c isomut2_lib.o fisher.o -lm -W -Wall'
__compiled.append(__subprocess.call(__cmd, shell=True))
__cmd = 'gcc -O3 -o checkPileup checkPileup.c isomut2_lib.o fisher.o -lm -W -Wall'
__compiled.append(__subprocess.call(__cmd, shell=True))
__os.chdir(cwd)

if 1 in __compiled:
    print('WARNING: C scripts were not compiled correctly, some functions will not work.')
    print('Try compiling manually to ' + HOME + '/C')
else:
    print('Done.')
